@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.029';
+our $VERSION = '0.030';
 
 use EV;
 use Time::HiRes;
@@ -187,15 +187,11 @@ sub request {
 }
 
 sub drop {
-    my ($self, $c) = @_;
+    my ($self, $c, $force_socket_close) = @_;
     my $conn_id = ref($c) eq 'HASH' ? $c->{id} : $c;
-    my $conn = $self->{connections}{$conn_id}
-      or return;
-
+    my $conn = $self->{connections}{$conn_id} or return;
     _register_in_timeline($conn, "dropping connection from pool") if exists $conn->{debug_or_timeline};
-    _set_completed_state($self, $conn_id) unless $conn->{state} == YAHC::State::COMPLETED();
-
-    delete $self->{connections}{$conn_id};
+    _set_completed_state($self, $conn_id, $force_socket_close) unless $conn->{state} == YAHC::State::COMPLETED();
     return $conn;
 }
 
@@ -223,8 +219,7 @@ sub socket_cache {
 ################################################################################
 
 sub yahc_terminal_error {
-    my $error = shift;
-    return int($error) & YAHC::Error::TERMINAL_ERROR() == YAHC::Error::TERMINAL_ERROR() ? 1 : 0;
+    return $_[0] & YAHC::Error::TERMINAL_ERROR() == YAHC::Error::TERMINAL_ERROR() ? 1 : 0;
 }
 
 sub yahc_reinit_conn {
@@ -363,7 +358,6 @@ sub _set_init_state {
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if exists $conn->{debug_or_timeline};
     _call_state_callback($self, $conn, 'init_callback') if exists $conn->{has_init_callback};
 
-    retry:
     _close_or_cache_socket($self, $conn, 1); # force connection close if any (likely not)
     my $watchers = _delete_watchers_but_lifetime_timer($self, $conn_id); # implicit stop of all watchers
 
@@ -374,7 +368,29 @@ sub _set_init_state {
     # otherwise we can get off-by-one error in yahc_conn_attempts_left
     my $attempt = ++$conn->{attempt};
     if ($attempt == 1 || !exists $conn->{request}{_backoff}) {
-        goto retry if _init_helper($self, $conn_id) == 1;
+        if (_init_helper($self, $conn_id) == 1) {
+            _register_in_timeline($conn, "do attempt on next EV iteration, (iteration=%d)", $self->{loop}->iteration)
+                if exists $conn->{debug_or_timeline};
+
+            # from EV docs:
+            # idle watcher call the callback when there are no other pending
+            # watchers of the same or higher priority. The idle watchers are
+            # being called once per event loop iteration - until stopped.
+            #
+            # so, what we do is we start idle watcher with priority 1 which is
+            # higher then 0 used by all IO watchers. As result, the callback
+            # will be called at the end of this iteration. And others if neccessary.
+
+            my $retry_watcher = $watchers->{retry} ||= $self->{loop}->idle_ns(_get_safe_wrapper($self, $conn, sub {
+                shift->stop; # stop this watcher, _set_init_state will start if neccessary
+                _register_in_timeline($conn, "time for new attempt (iteration=%d)", $self->{loop}->iteration)
+                    if exists $conn->{debug_or_timeline};
+                _set_init_state($self, $conn_id)
+            }));
+
+            $retry_watcher->priority(1);
+            $retry_watcher->start;
+        };
     } else {
         my $backoff_delay = eval { $conn->{request}{_backoff}->($conn) };
 
@@ -385,10 +401,10 @@ sub _set_init_state {
 
         $self->{loop}->now_update;
         _register_in_timeline($conn, "setting backoff_timer to %.3fs", $backoff_delay) if exists $conn->{debug_or_timeline};
-        $watchers->{backoff_timer} = $self->{loop}->timer($backoff_delay, 0, sub {
-            _register_in_timeline($conn, "backoff timer of %.3fs expired", $backoff_delay) if exists $conn->{debug_or_timeline};
+        $watchers->{backoff_timer} = $self->{loop}->timer($backoff_delay, 0, _get_safe_wrapper($self, $conn, sub {
+            _register_in_timeline($conn, "backoff timer of %.3fs expired, time for new attempt", $backoff_delay) if exists $conn->{debug_or_timeline};
             _set_init_state($self, $conn_id) if _init_helper($self, $conn_id) == 1;
-        });
+        }));
     }
 }
 
@@ -422,7 +438,9 @@ sub _init_helper {
 
         1;
     } or do {
-        yahc_conn_register_error($conn, YAHC::Error::CONNECT_ERROR(), "connection %d attempt failed: $@", $conn->{attempt});
+        my $error = $@ || 'zombie error';
+        $error =~ s/\s+$//o;
+        yahc_conn_register_error($conn, YAHC::Error::CONNECT_ERROR(), "connection attempt %d failed: %s", $conn->{attempt}, $error);
         return 1;
     };
 
@@ -442,7 +460,7 @@ sub _set_connecting_state {
     my $connecting_cb = _get_safe_wrapper($self, $conn, sub {
         my $sockopt = getsockopt($sock, SOL_SOCKET, SO_ERROR);
         if (!$sockopt) {
-            yahc_conn_register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to do getsockopt(): $!");
+            yahc_conn_register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to do getsockopt(): '%s' errno=%d", "$!", $!+0);
             _set_init_state($self, $conn_id);
             return;
         }
@@ -511,7 +529,7 @@ sub _set_ssl_handshake_state {
 
     if (!IO::Socket::SSL->start_SSL($fh, %options, SSL_startHandshake => 0)) {
         return _set_user_action_state($self, $conn_id, YAHC::Error::SSL_ERROR() | YAHC::Error::TERMINAL_ERROR(),
-            "failed to start SSL session: $IO::Socket::SSL::SSL_ERROR");
+            sprintf("failed to start SSL session: %s", _format_ssl_error()));
     }
 
     my $handshake_cb = _get_safe_wrapper($self, $conn, sub {
@@ -526,7 +544,7 @@ sub _set_ssl_handshake_state {
             return $w->events(EV::WRITE) if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_WRITE;
         }
 
-        yahc_conn_register_error($conn, YAHC::Error::SSL_ERROR(), "Failed to complete SSL handshake: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
+        yahc_conn_register_error($conn, YAHC::Error::SSL_ERROR(), "Failed to complete SSL handshake: %s", _format_ssl_error());
         _set_init_state($self, $conn_id);
     });
 
@@ -553,7 +571,8 @@ sub _set_write_state {
     warn "YAHC: HTTP message has UTF8 flag set! This will result in poor performance, see docs for details!"
         if utf8::is_utf8($buf);
 
-    _register_in_timeline($conn, "sending request of $length bytes") if exists $conn->{debug_or_timeline};
+    _register_in_timeline($conn, "writing body of %d bytes\n%s", $length, ($length > 1024? substr($buf, 0, 1024) . '... (cut to 1024 bytes)' : $buf))
+        if exists $conn->{debug_or_timeline};
 
     my $write_cb = _get_safe_wrapper($self, $conn, sub {
         my $w = shift;
@@ -566,13 +585,12 @@ sub _set_write_state {
                     return $w->events(EV::WRITE) if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_WRITE;
                 }
 
-                yahc_conn_register_error($conn, YAHC::Error::WRITE_ERROR() | YAHC::Error::SSL_ERROR(),
-                                "Failed to send HTTPS data: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
+                yahc_conn_register_error($conn, YAHC::Error::WRITE_ERROR() | YAHC::Error::SSL_ERROR(), "Failed to send HTTPS data: %s", _format_ssl_error());
                 return _set_init_state($self, $conn_id);
             }
 
             return if $! == EWOULDBLOCK || $! == EINTR || $! == EAGAIN;
-            yahc_conn_register_error($conn, YAHC::Error::WRITE_ERROR(), "Failed to send HTTP data: $!");
+            yahc_conn_register_error($conn, YAHC::Error::WRITE_ERROR(), "Failed to send HTTP data: '%s' errno=%d", "$!", $!+0);
             _set_init_state($self, $conn_id);
         } elsif ($wlen == 0) {
             yahc_conn_register_error($conn, YAHC::Error::WRITE_ERROR(), "syswrite returned 0");
@@ -619,13 +637,12 @@ sub _set_read_state {
                     return $w->events(EV::WRITE) if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_WRITE;
                 }
 
-                yahc_conn_register_error($conn, YAHC::Error::READ_ERROR() | YAHC::Error::SSL_ERROR(),
-                                "Failed to receive HTTPS data: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
+                yahc_conn_register_error($conn, YAHC::Error::READ_ERROR() | YAHC::Error::SSL_ERROR(), "Failed to receive HTTPS data: %s", _format_ssl_error());
                 return _set_init_state($self, $conn_id);
             }
 
             return if $! == EWOULDBLOCK || $! == EINTR || $! == EAGAIN;
-            yahc_conn_register_error($conn, YAHC::Error::READ_ERROR(), "Failed to receive HTTP data: $!");
+            yahc_conn_register_error($conn, YAHC::Error::READ_ERROR(), "Failed to receive HTTP data: '%s' errno=%d", "$!", $!+0);
             _set_init_state($self, $conn_id);
         } elsif ($rlen == 0) {
             if ($content_length > 0) {
@@ -758,9 +775,12 @@ sub _set_completed_state {
     # this's a terminal state,
     # so setting this state should *NEVER* fail
     delete $self->{callbacks}{$conn_id};
-    my $conn = $self->{connections}{$conn_id}
-      or warn "YAHC: try to _set_completed_state() for unknown connection $conn_id",
+    my $conn = delete $self->{connections}{$conn_id};
+
+    if (!defined $conn) {
+        delete($self->{watchers}{$conn_id}), # implicit stop of all watchers
         return;
+    }
 
     $conn->{state} = YAHC::State::COMPLETED();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if exists $conn->{debug_or_timeline};
@@ -775,15 +795,16 @@ sub _build_socket_and_connect {
     my ($ip, $port) = @_;
 
     my $sock;
-    socket($sock, PF_INET, SOCK_STREAM, 0) or die "Failed to construct TCP socket: $!\n";
+    socket($sock, PF_INET, SOCK_STREAM, 0)
+        or die sprintf("Failed to construct TCP socket: '%s' errno=%d\n", "$!", $!+0);
 
-    my $flags = fcntl($sock, F_GETFL, 0) or die "Failed to get fcntl F_GETFL flag: $!\n";
-    fcntl($sock, F_SETFL, $flags | O_NONBLOCK) or die "Failed to set fcntl O_NONBLOCK flag: $!\n";
+    my $flags = fcntl($sock, F_GETFL, 0) or die sprintf("Failed to get fcntl F_GETFL flag: '%s' errno=%d\n", "$!", $!+0);
+    fcntl($sock, F_SETFL, $flags | O_NONBLOCK) or die sprintf("Failed to set fcntl O_NONBLOCK flag: '%s' errno=%d\n", "$!", $!+0);
 
     my $ip_addr = inet_aton($ip) or die "Invalid IP address";
     my $addr = pack_sockaddr_in($port, $ip_addr);
     if (!connect($sock, $addr) && $! != EINPROGRESS) {
-        die "Failed to connect: $!\n";
+        die sprintf("Failed to connect: '%s' errno=%d\n", "$!", $!+0);
     }
 
     return $sock;
@@ -796,7 +817,7 @@ sub _get_next_target {
     # TODO STATE_RESOLVE_DNS
     ($host, $port) = ($1, $2) if !$port && $host =~ m/^(.+):([0-9]+)$/o;
     $ip = $host if !$ip && $host =~ m/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/o;
-    $ip ||= inet_ntoa(gethostbyname($host) or die "Failed to resolve '$host': $!");
+    $ip ||= inet_ntoa(gethostbyname($host) or die "Failed to resolve $host\n");
     $scheme ||= $conn->{request}{scheme} || 'http';
     $port   ||= $conn->{request}{port} || ($scheme eq 'https' ? 443 : 80);
 
@@ -1062,6 +1083,8 @@ sub _log_message {
     printf STDERR "[%s.%-5d] [$$] $format\n", POSIX::strftime('%F %T', localtime($now)), $ms || 0, @_;
 }
 
+sub _format_ssl_error { return sprintf("'%s' errno=%d ssl_error='%s' ssl_errno=%d", "$!", 0+$!, "$IO::Socket::SSL::SSL_ERROR", 0+$IO::Socket::SSL::SSL_ERROR); }
+
 1;
 
 __END__
@@ -1492,7 +1515,10 @@ documentation L<https://metacpan.org/pod/IO::Socket::SSL>.
 =head2 drop
 
 Given connection HashRef or conn_id move connection to COMPLETED state (avoiding
-'USER ACTION' state) and drop it from internal pool.
+'USER ACTION' state) and drop it from internal pool. The function takes two
+parameters: first is either a connection id or connection HashRef. Second one
+is a boolean flag indicating whether connection's socket should closed or it
+might be reused.
 
 =head2 run
 
